#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include "debug.h"


#define DEBUG_MSG_BUFSIZE 256

struct str_val_table {
  const char *str;
  const int val;
};

struct str_val_table facility_map_table[] = {
  { .str = "user",   .val = LOG_USER },
  { .str = "kern",   .val = LOG_KERN },
  { .str = "mail",   .val = LOG_MAIL },
  { .str = "daemon", .val = LOG_DAEMON },
  { .str = "auth",   .val = LOG_AUTH },
  { .str = "lpr",    .val = LOG_LPR },
  { .str = "news",   .val = LOG_NEWS },
  { .str = "uucp",   .val = LOG_UUCP },
  { .str = "cron",   .val = LOG_CRON },
  { .str = "local0", .val = LOG_LOCAL0 },
  { .str = "local1", .val = LOG_LOCAL1 },
  { .str = "local2", .val = LOG_LOCAL2 },
  { .str = "local3", .val = LOG_LOCAL3 },
  { .str = "local4", .val = LOG_LOCAL4 },
  { .str = "local5", .val = LOG_LOCAL5 },
  { .str = "local6", .val = LOG_LOCAL6 },
  { .str = "local7", .val = LOG_LOCAL7 },
  { NULL, 0 }
};

/*! \brief Converts string contains syslog facility name to specific value
 *         LOG_*
 *  \param name - string contains facility name
 *  \param val  - pointer to value in which constant will be stored if name is
 *                valid
 *
 *  \return 0 in case of success, -1 otherwise
 *
 *  \sa     man 3 syslog
 */
static inline int facility_str2val( const char *name, int *val )
{
  const struct str_val_table *item = facility_map_table;

  int i;
  for ( i = 0; item[ i ].str != NULL; i++ ) {

    if( strcmp( name, item[ i ].str ) == 0 ) {
      *val = item[ i ].val;
      return 0;
    }
  }
  
  return -1;
}

static int dbglevel2priority( int level )
{
  switch( level ) {
    case 0:
      return LOG_ERR;

    case 1:
      return LOG_WARNING;

    case 2:
      return LOG_INFO;
  
    case 3:
    default:
      break;
  }

  return LOG_DEBUG;
} 

int log_init(log_desc_t *d, const log_opts_t *log_opts, const char *prefix)
{
#if 0
  printf("%s: level:%u dest:%s\n", __FUNCTION__ ,
                                     log_opts->loglevel, log_opts->logfile);
#endif

  // It is facility by default. It's will be rewrited very soon, but if an error 
  // will be detected before it, we should write to some log rather than crash 
  d->facility = LOG_USER;

  d->buf = malloc(DEBUG_MSG_BUFSIZE);
  if (!d->buf) {
    return -1;
  }

  if (!log_opts) {
    return -1;
  }

  snprintf(d->prefix, sizeof(d->prefix), "%s", prefix);
  char *c=d->prefix;
  while (*c) *c = toupper(*c), c++;

  d->file = 0;
  if ((log_opts->logfile) && (log_opts->logfile[0])) {
    
    if (strcmp(log_opts->logfile, "sys")) {
      /* Not syslog */

      d->file = fopen(log_opts->logfile, "at+");
      if (!d->file) {
        printf("Can't open log file '%s'\n", log_opts->logfile);
        goto error;
      }

    } else {
      /* Syslog */

      if (facility_str2val( log_opts->facil_name, &d->facility ) ) {
        printf("Wrong log facility '%s'\n", log_opts->facil_name );
        goto error;
      }

#if 0
      setlogmask (LOG_UPTO (LOG_NOTICE));
#endif
      openlog(d->prefix, LOG_CONS | LOG_PID | LOG_NDELAY, d->facility);
    }

  } else
    d->file = stdout;

  d->level = log_opts->loglevel;
  d->bufsize = DEBUG_MSG_BUFSIZE;

  return 0;

error:

  log_free(d);
  return 1;
}

/*! \brief Function that cleans fields of log_desc_t structure that describes 
 *         log behavior for a daemon
 *
 *  \param d -- log descriptor to free
 *
 *  \note function clears all fields of structure but don't free memory that was
 *        be allocated for 'd'. You can free it separately.
 */
void log_free(log_desc_t *d)
{
  if (d == NULL ) {
    return;
  }

  if ( d->file ) {
     if (d->file != stdout) {
       fclose(d->file);
     }
  } else {
    closelog();
  }

  if( d->buf ) {
    free(d->buf); 
    d->buf = NULL;
  }
}

void log_print(log_desc_t *d, int mlevel, const char *format, ...)
{
  va_list args;

  if ( !((d) && (d->buf)) )
    return;

  if ( d->level >= mlevel ) {

    va_start( args, format);
    vsnprintf(d->buf, DEBUG_MSG_BUFSIZE, format, args );
    va_end(args);

    if (d->file) {
      fprintf(d->file, "%s: %s",d->prefix, d->buf);
      fflush(d->file);
    } else {
      syslog( d->facility | dbglevel2priority( mlevel ), "%s", d->buf );
    }
  }
}
