#ifndef LIBDAEMON_H
#define LIBDAEMON_H
#include <stdint.h>
#include <stddef.h>
#include <sys/types.h>
#include <sys/un.h>

typedef struct {
  uint16_t len;
  uint16_t code;
} libdmn_msg_hdr_t;

typedef struct {
  libdmn_msg_hdr_t hdr;
  uint8_t status;
} libdmn_msg_response_hdr_t;

typedef struct {
  libdmn_msg_hdr_t hdr;
  char cmd[1];
} libdmn_msg_str_cmd_t;

typedef struct {
  libdmn_msg_response_hdr_t rhdr;
  char response[1];
} libdmn_msg_str_cmd_r_t;

typedef int (*libdmn_str_cmd_handler_t)(int, char **,  void *);

typedef struct {
  const char *cmd;
  const char *help;
  libdmn_str_cmd_handler_t handler;
} libdmn_str_cmd_tbl_t;

typedef struct {
  pid_t childs[24];  
  int childs_cnt;
} libdmn_childs_t;

typedef struct array_s {
  uint16_t isize;
  uint16_t maxcnt;
  uint8_t *head;
} array_t;

array_t *array_create(size_t isize, uint16_t maxcnt);
void array_free(array_t *a);
void *array_at(array_t *a, uint16_t inum);
uint16_t array_idx(array_t *a, void *item);
size_t array_isize(array_t *a);
uint16_t array_size(array_t *a);

typedef struct {
  array_t *a;
  uint16_t cnt;
} list_t;

list_t *list_create(size_t isize, uint16_t maxcnt);
void list_free(list_t *l);
int list_add(list_t *l, void *item);
int list_del(list_t *l, void *item);
int list_del_by_idx(list_t *l, uint16_t idx);
int list_find(list_t *l, void *);
uint16_t list_size(list_t *l);
uint16_t list_count(list_t *l);
void *list_at(list_t *l, uint16_t idx);
void *list_last(list_t *l);
void list_clear(list_t *l);
#define list_isize(l) array_isize(l->a)


#define ptrlist_create(maxcnt) list_create(sizeof(void*), maxcnt)
#define ptrlist_find(l, p) list_find(l, &(p))
#define ptrlist_add(l, p) list_add(l, &(p))
#define ptrlist_del(l, p) list_del(l, &(p))
#define ptrlist_at(l, i) (*(void **)list_at(l, i))
#define ptrlist_last(l) (*(void **)list_last(l))
void ptrlist_free_with_items(list_t *l);

typedef struct {
  int fd;
  uint8_t mode;
} fdlist_item_t;

typedef list_t fdlist_t;

#define fdlist_create(maxcnt) list_create(sizeof(fdlist_item_t), maxcnt)
#define fdlist_free(fdl) list_free(fdl);
int fdlist_add(fdlist_t *fdl, int fd, uint8_t modes);
int fdlist_del(fdlist_t *fdl, int fd);
#define fdlist_at(fdl, idx) ((fdlist_item_t *)list_at(fdl, idx))
#define fdlist_clear(fdl) list_clear(fdl)
#define fdlist_count(fdl) list_count(fdl)
#define fdlist_size(fdl) list_size(fdl)

typedef struct {
  void *dbg;
} libdmn_t;

struct cli_option_item_t;

typedef struct {
  char logfile[128];
  char facil_name[32];
  char prior_name[32]; // Priority is syslog variable, and this variable 
                       // contains it's string name
  uint16_t loglevel;   // but loglevel is local variable. It is maximum level of 
                       // message that we print to log. And these messages can
                       // have any priority 
} log_opts_t;

typedef struct {
  char name[32];
  char *description;
  uint16_t sleeptimer;
  uint16_t bufsize;
  log_opts_t log_opts;
  char cm[128];
  time_t starttime;
  uint8_t foreground;
  uint8_t help;

  list_t *cmd;    
  list_t *fdhandlers;
  list_t *timers;
  list_t *period_timers;
  struct cli_option_item_s **cl_opts;
  struct cli_option_item_s **cl_args;

  struct sockaddr_un sa;
  int fd;
  uint8_t *buf;
  fdlist_t *fdl, *fdlr, *mng_fdl;
} daemon_common_var_t;

typedef struct daemon_s {
  void *log;
  libdmn_t h;
  daemon_common_var_t cv;
  void *v;

  void (*cleanup)(struct daemon_s *d);
  void (*sig_handler)(int);
  char *(*get_name)();
  char *(*get_description)();
  size_t (*get_vars_size)();
  int (*init_vars)(struct daemon_s *d);
  int (*add_opts)(struct daemon_s *d);
  int (*init)(struct daemon_s *d);
  int (*reset)(struct daemon_s *d);  
} daemon_t;

typedef struct {
  int (*init)(daemon_t *d);
  int (*check)(daemon_t *d, int fd);
  int (*handle_rx)(daemon_t *d, int fd);
  int (*handle_close)(daemon_t *d, int fd);
} libdmn_fd_handler_item_t;


typedef int (*libdmn_timer_handler_t)(daemon_t *d, void *data);

typedef struct {
  uint32_t period;
  uint64_t started;
  void *data;
  libdmn_timer_handler_t handler;
} libdmn_timer_t;

/* new API */

/* Main loop */
int libdmn_daemon_run(daemon_t *d, int argc, char **argv);

/* FD handlers */
void libdmn_add_fd_handler(daemon_t *d, libdmn_fd_handler_item_t *i);
int libdmn_daemon_add_fd(daemon_t *d, int fd);
int libdmn_daemon_del_fd(daemon_t *d, int fd);

/* String commands */
int libdmn_add_cm_handler(daemon_t *d, uint16_t cmid, libdmn_str_cmd_handler_t *handler);
void libdmn_mk_str_cmd_response(daemon_t *d, uint8_t statuscode, const char *fmt, ...);
void libdmn_add_str_cmd_response_line(daemon_t *d, char **p, const char *fmt, ...);
void libdmn_add_str_cmd_response_status(daemon_t *d, uint8_t statuscode);
int libdmn_add_str_cmd_handler(daemon_t *d, const char *name, libdmn_str_cmd_handler_t *handler);
int libdmn_add_str_cmd_handlers(daemon_t *d, libdmn_str_cmd_tbl_t *table);

/* Timers */
int libdmn_timer_init(libdmn_timer_t *, libdmn_timer_handler_t handler, void *data);
int libdmn_timer_start(daemon_t *d, libdmn_timer_t *, uint32_t period);
int libdmn_timer_stop(daemon_t *d, libdmn_timer_t *);
uint32_t libdmn_timer_elapsed(libdmn_timer_t *);
void libdmn_add_periodical_handler(daemon_t *d, libdmn_timer_handler_t handler, uint16_t period);
uint64_t libdmn_gettime_ms();

/* Others */
int libdmn_handle_cmd_opts(daemon_t *d,
                           void *useroptions,
                           void *userargs,
                            int argc, char **argv);

int libdmn_add_opts(daemon_t *d, struct cli_option_item_s *opts);
void libdmn_daemon_sig_handler(daemon_t *d, int signum);
int libdmn_log2(daemon_t *d, int mlevel, const char *format, ...);


/* old api */
int libdmn_init(libdmn_t **h, const log_opts_t *log_opts, const char *prfx);
int libdmn_free(libdmn_t *);

/* IO */
int libdmn_waitfd(libdmn_t *h, int fd, uint16_t timer, char m);

int libdmn_read_message(libdmn_t *, int fd,  uint8_t *buf);
int libdmn_write_message(libdmn_t *, int fd,  uint8_t *buf);
int libdmn_process_str_cmd(libdmn_t *, void *rbuf, libdmn_str_cmd_tbl_t *cmd_tbl,
                           uint16_t cmd_count, void *data);

int libdmn_execute_cm(libdmn_t *, char *sockpath, void *buf, size_t bufsize, char *command);

/* Handle childs */
int libdmn_handle_event(libdmn_t *, libdmn_childs_t *ch, char *scriptfile, char *args);
void libdmn_del_child(libdmn_t *, libdmn_childs_t *ch , pid_t pid);
void libdmn_terminate_childs(libdmn_t *, libdmn_childs_t *ch);

int libdmn_setsighandler(libdmn_t *, int sig, void (*sig_handler)(int));

int libdmn_log(libdmn_t *, int mlevel, const char *format, ...);


#define LIBDMN_MSG_RESPONSE    0x8000
#define LIBDMN_MSG_STR_CMD     0x0000
#define LIBDMN_MSG_TERM_CMD    0x0001
#define LIBDMN_MSG_UNKNOWN     0xFF00

#define LIBDMN_MKRESPONSE(rhdr, ln, cod, sts) \
  ((libdmn_msg_response_hdr_t*)rhdr)->status = sts; \
  ((libdmn_msg_response_hdr_t*)rhdr)->hdr.code = htons(cod | LIBDMN_MSG_RESPONSE); \
  ((libdmn_msg_response_hdr_t*)rhdr)->hdr.len = htons(ln+1);

#define LIBDMN_MKSTRRESPONSE(hdr, sts) LIBDMN_MKRESPONSE( (&hdr->rhdr), strlen(hdr->response)+1, LIBDMN_MSG_STR_CMD, sts)


/* Some STL items */

typedef struct {
  uint8_t *head;
  uint16_t size;
  uint8_t bitsize;
} bitmap_t;

bitmap_t *create_bitmap(uint8_t bitsize, uint8_t cnt);
void free_bitmap(bitmap_t *b);
uint8_t get_bitmap_item(bitmap_t *b, uint16_t inum);
void set_bitmap_item(bitmap_t *b, uint16_t inum, uint8_t v);

#endif
